##############################################################################
# ThanCad 0.6.2 "Students2021": n-dimensional CAD with raster support for engineers
# 
# Copyright (C) 2001-2021 Thanasis Stamos, December 8, 2021
# Athens, Greece, Europe
# URL: http://thancad.sourceforge.net
# e-mail: cyberthanasis@excite.com
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (www.gnu.org/licenses/gpl.html).
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##############################################################################
"""\
ThanCad 0.6.2 "Students2021": n-dimensional CAD with raster support for engineers

This module defines objects which read spreadsheet files (.xls, .xlsx, .ods)
for points or lines and create the appropriate ThanCad's elements.
"""

try: import xlrd          #Hopefully py2exe and cxFreeze get the hint to include xlrd
except ImportError: pass
from p_gimdxf import ThanImportBase
from thantrans import T


class ThanImportXlsBase(ThanImportBase):
    "A class to read point coordinates from a spreadsheet (xls/xlsx) file."

    @staticmethod
    def isEmpty(sh, i, j):
        "Test if cell is empty."
        if sh.cell_type(i, j) == xlrd.XL_CELL_EMPTY: return True  #Empty cell
        if sh.cell_type(i, j) == xlrd.XL_CELL_BLANK: return True  #Blank text
        if sh.cell_type(i, j) == xlrd.XL_CELL_TEXT:
            return sh.cell_value(i, j).strip() == ""              #Blank text
        return False                                              #Number, date, boolean, error


    @staticmethod
    def toString(sh, i, j):
        "Transform content of cell to string."
        if sh.cell_type(i, j) == xlrd.XL_CELL_EMPTY: return ""  #Empty cell
        if sh.cell_type(i, j) == xlrd.XL_CELL_BLANK: return ""  #Blank text
        if sh.cell_type(i, j) == xlrd.XL_CELL_TEXT: return sh.cell_value(i, j) #text
        v = sh.cell_value(i, j)
        if v == int(v): v = int(v)   #Αν τα δεκαδικά είναι 0, είναι κατά 99.99% σίγουρο ότι είναι ακέραιος
        return str(v)                #Convert number to string


    @staticmethod
    def toFloat(sh, i, j):
        "Transform content of cell to real number."
        if sh.cell_type(i, j) == xlrd.XL_CELL_EMPTY: return 0.0  #Empty cell
        if sh.cell_type(i, j) == xlrd.XL_CELL_BLANK: return 0.0  #Blank text
        if sh.cell_type(i, j) == xlrd.XL_CELL_TEXT:
            try: return float(sh.cell_value(i, j))               #Convert text to number
            except: return None                                  #Conversion failed
        if sh.cell_type(i, j) == xlrd.XL_CELL_ERROR: return None #Error code
        return sh.cell_value(i, j) + 0.0                         #Convert int and boolean to float


    def openXls(self, minrows, mincols):
        "Open the .xls or .xlsx file, make some tests and return the book object."
        try:
            import xlrd
        except ImportError:
            self.thanEr2s(T["Can not import xls/xlsx spreadsheets: The xlrd library/package was not found.\n"\
                "Please install xlrd in your system and retry."])
        fn = self.fDxf.name
        prt = self.thanDr.prt
        prt("Opening {}...".format(fn), "info1")
        try:
            book = xlrd.open_workbook(fn)
        except BaseException as e:
            self.thanEr2s(str(e))
        prt(T["The number of {0} worksheets is {1}"].format(fn, book.nsheets), "info1")
        if book.nsheets > 1: prt(T["Warning: Only the first sheet will be read"], "can")
        sh = book.sheet_by_index(0)
        prt(T["Worksheet {0}: {1} rows, {2} columns"].format(sh.name, sh.nrows, sh.ncols), "info1")
        if sh.ncols < 1 or sh.nrows < 1:
            self.thanEr2s(T["Sheet {} is empty!"].format(sh.name))
        elif sh.ncols < mincols:
            self.thanEr2s(T["Sheet {} has less than {} columns!"].format(sh.name, mincols))
        elif sh.nrows < minrows:
            self.thanEr2s(T["Sheet {} has less than {} rows!"].format(sh.name, minrows))
        return book, sh


    def iteraxyz(self, sh):
        """Iterate through all the point coordinates in the xls file: aa, x, y, z.

        icod is 0 if a point was found, 1 if empty line was found or 2 if syntax
        error or empty cell."""
        ztoo = sh.ncols > 3
        for i in range(sh.nrows):
            self._lindxf = i+1
            empty = self.isEmpty(sh, i, 0) and self.isEmpty(sh, i, 1) and self.isEmpty(sh, i, 2)
            if ztoo: empty = empty and self.isEmpty(sh, i, 3)
            if empty:
                yield 1, None, None, None, None
            else:
                aa = self.toString(sh, i, 0)
                x = self.toFloat(sh, i, 1)
                y = self.toFloat(sh, i, 2)
                z = self.toFloat(sh, i, 3) if ztoo else 0.0
                icod = 2 if x is None or y is None or z is None else 0
                yield icod, aa, x, y, z


    def iterxyz(self, sh):
        """Iterate through all the point coordinates in the xls file: x, y, z.

        icod is 0 if a point was found, 1 if empty line was found or 2 if syntax
        error or empty cell."""
        ztoo = sh.ncols > 2
        for i in range(sh.nrows):
            self._lindxf = i+1
            empty = self.isEmpty(sh, i, 0) and self.isEmpty(sh, i, 1)
            if ztoo: empty = empty and self.isEmpty(sh, i, 2)
            if empty:
                yield 1, None, None, None
            else:
                x = self.toFloat(sh, i, 0)
                y = self.toFloat(sh, i, 1)
                z = self.toFloat(sh, i, 2) if ztoo else 0.0
                icod = 2 if x is None or y is None or z is None else 0
                yield icod, x, y, z


    def iterfloats(self, sh):
        """Iterate through all lines of xls file and convert all cell values to floats.

        icod is 0 if valid floats were found in a line, 1 if an empty line was found or
        2 if syntax error or empty cell."""
        for i in range(sh.nrows):
            self._lindxf = i+1
            if all(self.isEmpty(sh, i, j) for j in range(sh.ncols)): yield 1, None
            vs = [self.toFloat(sh, i, j) for j in range(sh.ncols)]
            if any(v is None for v in vs): yield 2, vs
            yield 0, vs


class ThanImportXlsPoints(ThanImportXlsBase):
    "A class to import (topographic) points from a spreadsheet (xls/xlsx) file."

    def thanImport(self):
        "Import points from excel file."
        book, sh = self.openXls(1, 3)
        np = 0
        handle = ""
        validc = [True, True, True]
        if sh.ncols < 4: validc[2] = False
        for icod, aa, x, y, z in self.iteraxyz(sh):
            if icod == 1:
                self.thanWarn(T["blank row is ignored."])
            elif icod == 2:
                self.thanWarn(T["x, y or z is not a number. Point is ignored."])
            else:
                self.thanDr.dxfPoint(x, y, z, self.defLay, handle, None, aa, validc)
                np += 1
        lay = self.thanDr._dr.thanLayerTree.thanFindic(self.defLay)
        if lay is not None:   #If self.defLay != "0" and no points were added, self.defLay has not been created
            lay.thanAtts["hidename"].thanValSet(False)
            if sh.ncols > 3: lay.thanAtts["hideheight"].thanValSet(False)
        if sh.ncols > 3:
            self.thanDr.prt(T["{} points imported (out of {} spreadsheet rows)"].format(np, sh.nrows), "info")
        else:
            self.thanDr.prt(T["{} points (with no z) imported (out of {} spreadsheet rows)"].format(np, sh.nrows), "info")
        del sh, book


class ThanImportXlsLines(ThanImportXlsBase):
    "A class to import lines from a spreadsheet (xls/xlsx) file."

    def thanImport(self):
        "Import points from excel file."
        book, sh = self.openXls(1, 2)
        handle = ""
        xx = []
        yy = []
        zz = []
        for icod, x, y, z in self.iterxyz(sh):
            print(icod, x, y, z)
            if icod == 1:
                if len(xx) > 0:
                    self.thanDr.dxfPolyline(xx, yy, zz, self.defLay, handle, None)
                    xx = []
                    yy = []
                    zz = []
            elif icod == 2:
                self.thanEr1s(T["x, y or z is not a number."])
            else:
                xx.append(x)
                yy.append(y)
                zz.append(z)
        if len(xx) > 0:
            self.thanDr.dxfPolyline(xx, yy, zz, self.defLay, handle, None)


class ThanImportXlsSurface(ThanImportXlsBase):
    """A class to import a surface from a spreadsheet (xls/xlsx) file.

    The surface is defined as z values of a grid with possibly variable step.
    The first row contains the x value of each column, and the first column
    contains the y values of each row.
    The surface is imported as a set of neiboring 3dfaces.
    """

    def thanImport(self):
        "Import points from excel file."
        handle = ""
        book, sh = self.openXls(3, 3)
        #First line contains the x coordinates of all columns
        it = self.iterfloats(sh)
        icod, xs = next(it)
        if icod == 1:
            self.thanEr1s(T["Blank row found."])
        elif icod == 2:
            if any(x is None for x in xs[1:]): #First cell may be blank or may contain anything
                self.thanEr1s(T["Some cells do not contain numbers."])
        #Second line contains the first set of z values
        icod, vsa = next(it)
        if icod == 1:
            self.thanEr1s(T["Blank row found."])
        elif icod == 2:
            self.thanEr1s(T["Some cells do not contain numbers."])
        #Iterate through 3rd line to end
        for icod, vsb in it:
            if icod == 1:
                self.thanEr1s(T["Blank row found."])
            elif icod == 2:
                self.thanEr1s(T["Some cells do not contain numbers."])
            ya = vsa[0]
            yb = vsb[0]
            for j in range(2, sh.ncols):
                xa = xs[j-1]
                xb = xs[j]
                xx = xa,       xb,     xb,     xa
                yy = ya,       ya,     yb,     yb
                zz = vsa[j-1], vsa[j], vsb[j], vsb[j-1]
                self.thanDr.dxf3dface(xx, yy, zz, self.defLay, handle, None)
            vsa = vsb
